# -*- coding: utf-8 -*-


from __future__ import absolute_import
from dp_tornado.engine.helper import Helper as dpHelper

import datetime


class PerformanceMeasurement(object):
    def __init__(self, log_prefix, enabled):
        self._log_prefix = log_prefix
        self._enabled = enabled

        if enabled:
            self._begin = datetime.datetime.now()
            self._lap = datetime.datetime.now()

    def lap(self, log_suffix='Lap', with_print=True):
        if not self._enabled:
            return

        lap = self._lap
        self._lap = datetime.datetime.now()

        diff = (self._lap - lap).total_seconds()

        if with_print:
            if diff < 0.0001:
                diff = '0.000100'

            print('%s Elapsed : %s secs (%s)' % (self._log_prefix, diff, log_suffix))

        return diff

    def stop(self, log_suffix='Stop', with_print=True):
        if not self._enabled:
            return

        self._stop = datetime.datetime.now()

        diff = (self._stop - self._begin).total_seconds()

        if with_print:
            if diff < 0.0001:
                diff = '0.000100'

            print('%s Elapsed : %s secs (%s)' % (self._log_prefix, diff, log_suffix))

        return diff


class PerformanceHelper(dpHelper):
    def start(self, log_prefix='Time', enabled=True):
        return PerformanceMeasurement(log_prefix, enabled)



# -*- coding: utf-8 -*-


from __future__ import absolute_import

import importlib


class Loader(object):
    def __init__(self, category=None, path_prefix=None):
        self.__dict__['_category'] = category
        self.__dict__['_path_prefix'] = path_prefix if path_prefix else category
        self.__dict__['_prepared'] = {}

    def __getattr__(self, name):
        return self.__getattr_inner__(name)

    def __getattr_inner__(self, name, engine=False):
        try:
            attr = self.__getattribute__(name)
        except AttributeError:
            attr = None

        if attr:
            return attr

        if name in self.__dict__:
            return self.__dict__[name]

        if name in self._prepared:
            return self._prepared[name]

        engine_prefix = 'dp_tornado.' if engine else ''
        module_path = '%s%s' % (engine_prefix, self.__dict__['_path_prefix'])

        try:
            module = importlib.import_module('%s.%s' % (module_path, name))
            obj = getattr(
                module, '%s%s' % (self.capitalized_method_name(name), self.__dict__['_category'].capitalize()))

        except ImportError:
            return None if engine else self.__getattr_inner__(name, True)

        except AttributeError:
            self._prepared[name] = Loader(self.__dict__['_category'], '%s.%s' % (module_path, name))
            return self._prepared[name]

        _prepared = obj()
        _prepared.__dict__['_category'] = self.__dict__['_category']
        _prepared.__dict__['_path_prefix'] = '%s.%s' % (self.__dict__['_path_prefix'], name)

        self._prepared[name] = _prepared

        if hasattr(_prepared, 'index'):
            if _prepared.index:
                _prepared.index()

        return self._prepared[name]

    @staticmethod
    def capitalized_method_name(method_name):
        s = method_name.split('_')
        x = map(lambda p: p.capitalize(), s)

        return ''.join(x)


# -*- coding: utf-8 -*-


from __future__ import absolute_import
from dp_tornado.engine.helper import Helper as dpHelper

import time
import datetime


class DatetimeHelper(dpHelper):
    def current_time(self):
        return self.helper.numeric.long(time.time())

    def time(self, dt=None):
        return self.current_time() if dt is None else self.timestamp_from_datetime(dt, False)

    def current_time_millis(self):
        return self.helper.numeric.long(round(time.time() * 1000))

    def mtime(self, dt=None):
        return self.current_time_millis() if dt is None else self.timestamp_from_datetime(dt, True)

    def today(self):
        return datetime.datetime.today()

    def yesterday(self, d=None):
        if isinstance(d, int) and 19700000 <= d <= 99999999:
            d = str(d)
            d = self.datetime(self.mktime(int(d[0:4]), int(d[4:6]), int(d[6:8])))

        elif d is not None:
            raise ValueError

        else:
            d = self.today()

        return d - self.timedelta(days=1)

    def tomorrow(self, d=None):
        if isinstance(d, int) and 19700000 <= d <= 99999999:
            d = str(d)
            d = self.datetime(self.mktime(int(d[0:4]), int(d[4:6]), int(d[6:8])))

        elif d is not None:
            raise ValueError

        else:
            d = self.today()

        return d + self.timedelta(days=1)

    def ago(self,
            days=0,
            seconds=0,
            microseconds=0,
            milliseconds=0,
            minutes=0,
            hours=0,
            weeks=0,
            ts=False,
            millisecs=False):
        ago = self.today() - self.timedelta(days, seconds, microseconds, milliseconds, minutes, hours, weeks)

        if ts:
            return self.timestamp_from_datetime(ago, millisecs=millisecs)

        else:
            return ago

    def yyyymmdd(self, s='', d=None):
        if not isinstance(d, datetime.datetime):
            d = self.datetime(d)

        return '%04d%s%02d%s%02d' % (d.year, s, d.month, s, d.day)

    def mmdd(self, s='', d=None):
        if not isinstance(d, datetime.datetime):
            d = self.datetime(d)

        return '%02d%s%02d' % (d.month, s, d.day)

    def hhiiss(self, s='', d=None):
        if not isinstance(d, datetime.datetime):
            d = self.datetime(d)

        return '%02d%s%02d%s%02d' % (d.hour, s, d.minute, s, d.second)

    def hhii(self, s='', d=None):
        if not isinstance(d, datetime.datetime):
            d = self.datetime(d)

        return '%02d%s%02d' % (d.hour, s, d.minute)

    def hour(self, d=None):
        if isinstance(d, datetime.datetime):
            return d.hour

        return self.datetime(d).hour

    def year(self, d=None):
        if isinstance(d, datetime.datetime):
            return d.year

        return self.datetime(d).year

    def month(self, d=None):
        if isinstance(d, datetime.datetime):
            return d.month

        return self.datetime(d).month

    def day(self, d=None):
        if isinstance(d, datetime.datetime):
            return d.day

        return self.datetime(d).day

    """
    isoweekday=True:  Mon=1, Tue=2, ..., Sun=7
    isoweekday=False: Mon=0, Tue=1, ..., Sun=6
    """
    def weekday(self, d=None, isoweekday=True):
        if isinstance(d, datetime.datetime):
            return d.isoweekday() if isoweekday else d.weekday()

        return self.datetime(d).isoweekday() if isoweekday else self.datetime(d).weekday()

    def datetime(self, ts=None, yyyymmdd=None, hhiiss='000000'):
        if yyyymmdd:
            yyyymmdd = int(yyyymmdd)

            if 19700000 <= yyyymmdd <= 99999999:
                yyyymmdd = str(yyyymmdd)
                hhiiss = str(hhiiss)
                y = int(yyyymmdd[0:4])
                m = int(yyyymmdd[4:6])
                d = int(yyyymmdd[6:8])
                h = int(hhiiss[0:2])
                i = int(hhiiss[2:4])
                s = int(hhiiss[4:6])
                v = self.datetime(self.mktime(y, m, d, h, i, s))

                if v.year == y and v.month == m and v.day == d:
                    return v

            raise ValueError

        elif ts:  # from timestamp
            ts = self.helper.numeric.long(ts)

            if ts > 9999999999:  # with microseconds
                return datetime.datetime.fromtimestamp(ts // 1000).replace(microsecond=ts % 1000 * 1000)
            else:
                return datetime.datetime.fromtimestamp(ts)
        else:
            return datetime.datetime.now()

    def millisecs_to_next_hour(self, from_time=None):
        dst = self.today() + self.timedelta(hours=1)
        from_time = from_time if from_time is not None else self.current_time_millis()

        return self.mktime(dst.year, dst.month, dst.day, dst.hour, 0, 0, millisecs=True) - from_time

    def secs_to_next_hour(self, from_time=None):
        return self.helper.numeric.long(self.millisecs_to_next_hour(from_time=from_time) / 1000)

    def millisecs_to_tomorrow(self, from_time=None):
        dst = self.today() + self.timedelta(days=1)
        from_time = from_time if from_time is not None else self.current_time_millis()

        return self.mktime(dst.year, dst.month, dst.day, 0, 0, 0, millisecs=True) - from_time

    def secs_to_tomorrow(self, from_time=None):
        return self.helper.numeric.long(self.millisecs_to_tomorrow(from_time=from_time) / 1000)

    def timedelta(self, days=0, seconds=0, microseconds=0, milliseconds=0, minutes=0, hours=0, weeks=0):
        return datetime.timedelta(days, seconds, microseconds, milliseconds, minutes, hours, weeks)

    def mktime(self, year=1970, month=1, day=1, hour=0, mins=0, sec=0, millisecs=False, utctimetuple=None):
        ts = self.helper.numeric.long(
            time.mktime(utctimetuple if utctimetuple else (year, month, day, hour, mins, sec, 0, 0, 0)))

        return ts * (1000 if millisecs else 1)

    def timestamp_from_datetime(self, dt, millisecs=False):
        if not isinstance(dt, datetime.datetime):
            return None

        if getattr(dt, 'timestamp', None):
            return self.helper.numeric.long(dt.timestamp() * (1000 if millisecs else 1))
        else:
            return self.mktime(utctimetuple=dt.utctimetuple(), millisecs=millisecs)



# -*- coding: utf-8 -*-


import os
import requests
import tornado.web
import tornado.options
import subprocess
import tempfile

from ..cache import dpInValueModelConfig
from ..engine import Engine as dpEngine


class Compressor(dpEngine):
    def __init__(self, handler):
        self.handler = handler
        self.provider = handler.settings.get('compressor', False) if handler else None

        if not self.provider:
            raise Exception('No compressor provider specified.')

        if not self.provider['yuicompressor']:
            raise Exception('No YUICompress location specified.')

        if not self.provider['uglifyjs']:
            raise Exception('No UglifyJS location specified.')

    def already_compressed(self, location):
        for k in ('-min-', '-min.', '.min.', '.minified.', '.pack.', '-jsmin.'):
            if k in location:
                return True

        return False

    def compress(self, location):
        if self.already_compressed(location):
            return self._read(location)

        c = self._compressor(location)

        if not c:
            raise Exception('Not supported extension.')

        try:
            c = self._compress(c, location)
            return c
        except Exception as e:
            self.logger.exception(e)
            return self._read(location)

    def _compressor(self, location, options=''):
        if location.endswith('.js'):
            return '%s %s %s' % (self.provider['uglifyjs'], options, location)
        elif location.endswith('.css'):
            return 'java -jar %s %s %s' % (self.provider['yuicompressor'], options, location)

    def _compress(self, command, location):
        p = subprocess.Popen(
            command, shell=True, stdout=subprocess.PIPE, stdin=subprocess.PIPE, stderr=subprocess.PIPE)

        try:
            (stdout, stderr) = p.communicate()
        except OSError as msg:
            raise Exception('An error has occurred from OSError. (%s)' % msg)

        if stderr:
            raise Exception('An error has occurred from compressor. (%s)' % stderr.decode())

        if self.helper.system.py_version <= 2:
            return stdout
        else:
            return stdout.decode()

    def _read(self, location):
        f = open(location, 'r')
        r = f.readlines()
        f.close()

        return ''.join(r)

    @staticmethod
    def clear(combined_dir):
        if not os.path.isdir(combined_dir):
            os.makedirs(combined_dir)

        for f in os.listdir(combined_dir):
            path = os.path.join(combined_dir, f)

            if not os.path.isdir(path):
                os.remove(path)


class StaticURL(tornado.web.UIModule):
    def render(self, *statics, **options):
        if not self.handler.vars.compressor:
            self.handler.vars.compressor = Compressor(self.handler)
            self.handler.vars.static.path = self.handler.settings.get('static_path')
            self.handler.vars.static.prefix = self.handler.settings.get('static_url_prefix')
            self.handler.vars.static.combined_path = self.handler.settings.get('combined_static_path')
            self.handler.vars.static.combined_prefix = self.handler.settings.get('combined_static_url_prefix')
            self.handler.vars.static.cache_config = dpInValueModelConfig('sqlite', 'static_url', pure=True)

            self.handler.vars.static.aws_id = tornado.options.options.static_aws_id
            self.handler.vars.static.aws_secret = tornado.options.options.static_aws_secret
            self.handler.vars.static.aws_bucket = tornado.options.options.static_aws_bucket
            self.handler.vars.static.aws_endpoint = tornado.options.options.static_aws_endpoint

            self.handler.vars.static.aws_configured = True if (self.handler.vars.static.aws_id and
                                                               self.handler.vars.static.aws_secret and
                                                               self.handler.vars.static.aws_bucket and
                                                               self.handler.vars.static.aws_endpoint) else False

        # for Debugging Mode, do not minify css or js.
        if self.handler.settings.get('debug'):
            return '\n'.join(
                [self._template('%s?%s' % (t, self.handler.vars.compressor.helper.datetime.mtime())) for t in statics])
        elif (options and 'proxy' in options and options['proxy']) or not self.handler.vars.static.aws_configured:
            return '\n'.join([self._template('%s?%s' % (t, self.handler.application.startup_at)) for t in statics])

        cache_key = 'key_%s_%s' % (len(statics), self.handler.helper.crypto.sha224_hash('/'.join(sorted(statics))))
        prepared = (self.handler.vars.static.prepared.__getattr__(cache_key) or
                    self.handler.cache.get(cache_key, dsn_or_conn=self.handler.vars.static.cache_config))

        if prepared:
            return prepared

        if 'separate' not in options or not options['separate']:
            extensions = {}

            for static in statics:
                ext = static.split('?')[0].split('.')[-1].lower()

                if ext not in extensions:
                    extensions[ext] = []

                extensions[ext].append(static)

            if not extensions:
                return ''

            if len(extensions) > 1:
                options['separate'] = False
                return '\n'.join(self.render(*static, **options) for static in list(extensions.values()))
            else:
                statics = list(extensions.values())[0]

        template = self._template(statics[0], replaced=False)
        ext = statics[0].split('.')[-1].lower()

        if not template:
            return ''

        acquire_key = 'job/%s' % cache_key
        acquire_identifier = self.handler.helper.random.uuid()

        if not self.handler.cache.get(acquire_key, dsn_or_conn=self.handler.vars.static.cache_config):
            self.handler.cache.set(
                acquire_key,
                acquire_identifier,
                dsn_or_conn=self.handler.vars.static.cache_config,
                expire_in=60*3)
        else:
            options['proxy'] = True
            return self.render(*statics, **options)

        fp = tempfile.NamedTemporaryFile(mode='w', suffix='.%s' % ext, delete=False)
        filename = '%s.%s' % (self.handler.helper.crypto.sha224_hash(self.handler.helper.random.uuid()), ext)
        content_length = 0

        for static in statics:
            compiled = self.handler.vars.compressor.compress(os.path.join(self.handler.vars.static.path, static))
            fp.write(compiled)
            content_length += len(compiled)

        s3bridge = self.handler.helper.aws.s3.connect(
            self.handler.vars.static.aws_id, self.handler.vars.static.aws_secret)

        fp_name = fp.name

        fp.close()
        fp = open(fp_name, 'r')

        s3bridge.set_contents_from_file(self.handler.vars.static.aws_bucket, filename, fp)

        fp.close()
        os.remove(fp_name)

        prepared_url = '%s/%s' % (self.handler.vars.static.aws_endpoint, filename)
        prepared = template % {'url': prepared_url}

        test = requests.get(prepared_url)

        if self.handler.cache.get(acquire_key, dsn_or_conn=self.handler.vars.static.cache_config) == acquire_identifier:
            if test.status_code != 200 or (content_length and not len(test.text.strip())):
                options['proxy'] = True
                return self.render(*statics, **options)

            self.handler.vars.static.prepared.__setattr__(cache_key, prepared)
            self.handler.cache.set(cache_key, prepared, dsn_or_conn=self.handler.vars.static.cache_config)

        return prepared

    def _template(self, path, replaced=True):
        if replaced:
            url = self.handler.vars.compressor.helper.url.urlparse.urljoin(self.handler.vars.static.prefix, path)
            return self._template(path, replaced=False) % {'url': url}

        extension = path.split('?')[0].split('.')[-1].lower()

        if extension == 'css':
            return '<link rel="stylesheet" type="text/css" href="%(url)s" />'
        elif extension == 'js':
            return '<script type="text/javascript" src="%(url)s"></script>'
        else:
            raise NotImplementedError
